@@ -6,6 +6,7 @@ import aiohttp
 import time
 import itertools
 import more_itertools
+import pickle
 
 from tqdm import tqdm
 
@@ -48,14 +49,15 @@ class SemaphoreWrapper():
     '''
 
 
-    def __init__(self, pool_size: int, cautious: bool):
+    def __init__(self, pool_size: int, requests_per_second: float, 
+                 cautious: bool):
         if cautious:
             self._stopped_time = time.monotonic()
             self._stopped_value = 0
         else:
             self._stopped_time = None
             self._stopped_value = None
-        self._REQUESTS_PER_SECOND = 2
+        self.requests_per_second = requests_per_second
         self._pool_size = pool_size
 
     async def __aenter__(self):
@@ -72,7 +74,7 @@ class SemaphoreWrapper():
             time_passed = time.monotonic() - self._stopped_time
 
             # сколько шагов должно было пройти
-            add_steps = time_passed / self._REQUESTS_PER_SECOND // 1
+            add_steps = time_passed / self.requests_per_second // 1
 
             # сколько шагов могло пройти с учетом ограничений + еще один
             real_add_steps = min(self._pool_size - self._stopped_value,
@@ -82,7 +84,7 @@ class SemaphoreWrapper():
             self._sem._value += real_add_steps
 
             # ждем время, излишне списанное при добавлении дополнительного шага
-            await asyncio.sleep((add_steps + 1) / self._REQUESTS_PER_SECOND - time_passed)
+            await asyncio.sleep((add_steps + 1) / self.requests_per_second - time_passed)
 
             self._stopped_time = None
             self._stopped_value = None
@@ -98,7 +100,7 @@ class SemaphoreWrapper():
         while True:
             if self._sem._value < self._sem._bound_value:
                 self._sem.release()
-            await asyncio.sleep(1 / self._REQUESTS_PER_SECOND)
+            await asyncio.sleep(1 / self.requests_per_second)
 
     async def acquire(self):
         '''
@@ -131,11 +133,15 @@ class Bitrix:
     - webhook: str - URL вебхука, полученного от сервера Битрикс
     
     - custom_pool_size: int = 50 - размер пула запросов. По умолчанию 50 запросов - 
-    это размер, указанный в официальной документации Битрикс24
-    на июль 2020 г. (https://dev.1c-bitrix.ru/rest_help/rest_sum/index.php)
+        это размер, указанный в официальной документации Битрикс24
+        на июль 2020 г. (https://dev.1c-bitrix.ru/rest_help/rest_sum/index.php)
     
+    - requests_per_second: float = 2 - скорость отправки запросов. По умолчанию
+        2 запроса в секунду - предельная скорость, согласно официальной
+        документации.
+
     - cautious: bool = False - стартовать, считая, что пул запросов уже
-    исчерпан, и нужно контролировать скорость запросов с первого запроса
+        исчерпан, и нужно контролировать скорость запросов с первого запроса
     
     - autobatch: bool = True - автоматически объединять списки запросов в батчи
 
@@ -145,17 +151,19 @@ class Bitrix:
     - get_all(self, method: str, details=None)
     - get_by_ID(self, method: str, ID_list, details=None)
     - post(self, method: str, item_list)
+    - set_requests_per_second(self, requests_per_second: float)
     '''
 
-    def __init__(self, webhook: str, custom_pool_size=50, cautious=False, 
-            autobatch=True, verbose=True):
+    def __init__(self, webhook: str, custom_pool_size: int = 50,
+                 requests_per_second: float = 2, cautious: bool = False,
+                 autobatch: bool = True, verbose: bool = True):
         '''
         Создает объект класса Bitrix.
 
         '''
         
-        self.webhook = webhook
-        self._sw = SemaphoreWrapper(custom_pool_size, cautious)
+        self.webhook = _correct_webhook(webhook)
+        self._sw = SemaphoreWrapper(custom_pool_size, requests_per_second, cautious)
         self._autobatch = autobatch
         self._verbose = verbose
 
@@ -177,7 +185,7 @@ class Bitrix:
 
             batch_size = 50
             while True:
-                batch = [{
+                batches = [{
                     'halt': 0,
                     'cmd': {
                         item['ID'] if preserve_IDs else f'cmd{i}': 
@@ -187,7 +195,7 @@ class Bitrix:
                     for next_batch in more_itertools.chunked(item_list, batch_size)
                 ]
                 uri_len = len(self.webhook + 'batch' +
-                              urllib.parse.urlencode(batch[0]))
+                              urllib.parse.urlencode(batches[0]))
                 if uri_len > BITRIX_URI_MAX_LEN:
                     batch_size = int(
                         batch_size // (uri_len / BITRIX_URI_MAX_LEN))
@@ -195,7 +203,7 @@ class Bitrix:
                     break
 
             method = 'batch'
-            item_list = batch
+            item_list = batches
 
         async with self._sw, aiohttp.ClientSession(raise_for_status=True) as session:
             tasks = [asyncio.create_task(self._request(session, method, i))
@@ -229,27 +237,21 @@ class Bitrix:
                 params.update({'order': {'ID': 'ASC'}})
         else:
             params = {'order': {'ID': 'ASC'}}
+
         async with self._sw, aiohttp.ClientSession(raise_for_status=True) as session:
             results, total = await self._request(session, method, params)
             if not total or total <= 50:
                 return results
-            remaining_results = await self._request_list(method, [
+
+            results.extend(await self._request_list(method, [
                 _merge_dict({'start': start}, params)
                 for start in range(len(results), total, 50)
-            ], total, len(results))
+            ], total, len(results)))
 
-        # дедуплицируем по id
-        dedup_results = results
-        for r in remaining_results:
-            if r['ID'] not in [dr['ID'] for dr in dedup_results]:
-                dedup_results.append(r)
+        # дедупликация через сериализацию, превращение в set и десериализацию
+        return ([pickle.loads(y) for y in set([pickle.dumps(x) for x in results])]  
+            if results else [])
 
-#       а более элегантный механизм ниже (через построение set()) не работает,
-#       так как результаты содержат вложенные списки, которые не хэшируются
-#       results = [dict(t) for t in {
-#            tuple(d.items()) for d in list(itertools.chain(results, remaining_results))
-#       }]
-        return dedup_results
 
     def get_all(self, method: str, params=None):
         '''
@@ -262,12 +264,19 @@ class Bitrix:
         Параметры:
         - method - метод REST API для запроса к серверу
         - params - параметры для передачи методу. Используется именно тот формат,
-            который указан в документации к REST API Битрикс24
+            который указан в документации к REST API Битрикс24. get_all() не
+            поддерживает параметры 'start', 'limit' и 'order'.
 
         Возвращает полный список сущностей, имеющихся на сервере,
         согласно заданным методу и параметрам.
         '''
 
+        if params:
+            _check_params(params)
+            for k in params.keys():
+                if k in ['start', 'limit', 'order']:
+                    raise ValueError("get_all() doesn't support parameters 'start', 'limit' or 'order'")
+
         return asyncio.run(self._get_paginated_list(method, params))
 
     def get_by_ID(self, method: str, ID_list, params=None):
@@ -298,6 +307,9 @@ class Bitrix:
         сущности.
         '''
 
+        if params: _check_params(params)
+        if len(ID_list) == 0:
+            return []
         return asyncio.run(self._request_list(
             method,
             [_merge_dict({'ID': ID}, params) for ID in ID_list] if params else
@@ -315,9 +327,31 @@ class Bitrix:
 
         Возвращает список ответов сервера для каждого из элементов item_list.
         '''
+        if len(item_list) == 0:
+            return []
+        try:
+            [_check_params(p) for p in item_list]
+        except (TypeError, ValueError) as err:
+            raise ValueError(
+                'item_list contains items with incorrect method params') from err 
         return asyncio.run(self._request_list(method, item_list))
 
 
+    def set_requests_per_second(self, requests_per_second: float) -> None:
+        '''
+        Установить скорость запросов к серверу Битрикса, равную 
+        requests_per_second. Может использоваться для понижения скорости
+        запросов припроведении большого количества операций, нагружающих сервер
+        (например, создание лидов или сделок), из-за чего он может возвращать
+        ошибку 500 Internal Server Error.
+
+        Параметры:
+        - requests_per_second - новая скорость в запросах в секунду
+
+        Возвращает None.
+        '''
+        self._sw.requests_per_second = requests_per_second
+
 ##########################################
 #
 #   internal functions
@@ -338,7 +372,8 @@ def _bitrix_url(data):
         return outStr
 
     def r_urlencode(data):
-        if isinstance(data, list) or isinstance(data, tuple):
+        if any(isinstance(data, t) for t in [list, tuple, set]):
+            data = list(data)
             for i in range(len(data)):
                 parents.append('[]')
                 r_urlencode(data[i])
@@ -360,3 +395,50 @@ def _merge_dict(d1, d2):
     if d2:
         d3.update(d2)
     return d3
+
+def _check_params(p):
+
+    # check if p is dict
+    if not isinstance(p, dict):
+        raise TypeError('params agrument should be a dict')
+
+    # check for allowed keys
+    clauses = {
+        'select': list,
+        'halt': int,
+        'cmd': dict,
+        'limit': int,
+        'order': dict,
+        'filter': dict,
+        'start': int,
+        'fields': dict
+    }
+
+    for pk in p.keys():
+        if pk not in clauses.keys():
+            raise ValueError(f'Unknown clause "{pk}" in params argument')
+
+    # check for allowed types of key values
+    for pi in p.items():
+        t = clauses[pi[0]]
+        if not (
+            (isinstance(pi[1], t)) or
+            ((t == list) and (any([isinstance(pi[1], x) for x in [list, tuple, set]])))
+        ):
+            raise TypeError(f'Clause "{pi[0]}" should be of type {t}, but its type is {type(pi[1])}')
+
+
+def _url_valid(url):
+    try:
+        result = urllib.parse.urlparse(url)
+        return all([result.scheme, result.netloc, result.path])
+    except:
+        return False
+
+
+def _correct_webhook(wh):
+    if not isinstance(wh, str):
+        raise TypeError(f'Webhook should be a {str}')
+    if not _url_valid(wh):
+        raise ValueError('Webhook is not a valid URL')
+    return wh if wh[-1] == '/' else wh + '/'
